//
// Created by tim on 05.04.23.
//

#include <fstream>
#include "CTUApexConsumer.h"
#include "Util.h"
#include "llvm/Demangle/Demangle.h"
#include "FunctionFinderVisitor.h"

void CTUApexConsumer::outlineFunctions() {
    for (const auto &node: call_graph->getNodes()) {
        if (auto md = node.second->checkAndGet<ExtractionMetadata>(); md.first && md.second->get_decl_node()) {
            ExtractionBlock eb;
            clang::SourceRange sourceRange = md.second->get_decl_node()->getSourceRange();
            eb.begin = sourceRange.getBegin();
            eb.end = sourceRange.getEnd();
            if(md.second->isMarkedToExtract()){
                //std::cout<<"Need to extract: "<<md.second->get_decl_node()->getNameAsString()<<"\n";
                eb.contents = clang::Lexer::getSourceText(clang::CharSourceRange::getTokenRange(sourceRange),
                                                          md.second->get_decl_node()->getASTContext().getSourceManager(),
                                                          clang::LangOptions(),
                                                          nullptr).str();
            }else{
                //std::cout<<"Not extracting: "<<md.second->get_decl_node()->getNameAsString()<<"\n";
                eb.contents="";
            }
            funcs.push_back(eb);
        }
    }

    //the callgraph nodes are unsorted, we need to sort the extraction blocks afterward
    std::sort(funcs.begin(),funcs.end(),[](const ExtractionBlock& a, const ExtractionBlock& b){return a.begin<b.begin;});
}

void CTUApexConsumer::outlineGlobals() {
    //Fixme: does this need to extract cross TU as well?
    SPDLOG_INFO("Compute set of used globals");
    std::unordered_set<clang::Decl *> availableGlobals;
    std::vector<ExtractionBlock> returnBlocks;
    for (const auto &node: call_graph->getNodes()) {
        if (auto md = node.second->checkAndGet<ExtractionMetadata>(); md.first && md.second->get_decl_node()) {
            assert(clang::isa<clang::FunctionDecl>(md.second->get_decl_node()) &&
                   "Extractionset should only contain function decls at this point");
            GlobalsMarker gMarker;
            auto globalsSet = gMarker.get_nodes_to_outline(clang::cast<clang::FunctionDecl>(md.second->get_decl_node()));
            for (const auto &func: globalsSet) {
                ExtractionBlock eb;
                eb.begin = func->getBeginLoc();
                eb.end = func->getEndLoc();
                clang::SourceRange range(eb.begin, eb.end);
                clang::CharSourceRange charSourceRange(range, true);
                eb.contents = clang::Lexer::getSourceText(charSourceRange, func->getASTContext().getSourceManager(),
                                                          clang::LangOptions(),
                                                          nullptr).str();
                globals.push_back(eb);
            }
        }
    }
}

ExtractionBlock CTUApexConsumer::getNextExtractionBlock() {
    //find the earliest block by starting source location
    ExtractionBlock min;
    if (!includes.empty()) {
        SPDLOG_TRACE("Setting include as assumed min");
        min = includes.front();
    } else if (!funcs.empty()) {
        SPDLOG_TRACE("Setting funcs as assumed min");
        min = funcs.front();
    } else if (!globals.empty()) {
        SPDLOG_TRACE("Setting globals as assumed min");
        min = globals.front();
    } else if (!macros.empty()) {
        SPDLOG_TRACE("Setting macros as assumed min");
        min = macros.front();
    } else if (!pragmas.empty()) {
        SPDLOG_TRACE("Setting pragmas as assumed min");
        min = pragmas.front();
    }


    while (!funcs.empty() && min.begin > funcs.front().begin) {
        SPDLOG_TRACE("Overwriting with funcs as min");
        min = funcs.front();
    }
    while (!globals.empty() && min.begin > globals.front().begin) {
        SPDLOG_TRACE("Overwriting with globals as min");
        min = globals.front();
    }
    while (!macros.empty() && min.begin > macros.front().begin) {
        SPDLOG_TRACE("Overwriting with macros as min");
        min = macros.front();
    }
    while (!pragmas.empty() && min.begin > pragmas.front().begin) {
        SPDLOG_TRACE("Overwriting with pragmas as min");
        min = pragmas.front();
    }

    //if the earliest blocks end is after another block, we fully contain the other block
    //we don't need to merge that block anymore
    while (!includes.empty() && min.end >= includes.front().end) {
        SPDLOG_TRACE("Erasing include entry");
        includes.erase(includes.begin());
    }
    while (!funcs.empty() && min.end >= funcs.front().end) {
        SPDLOG_TRACE("Erasing funcs entry");
        funcs.erase(funcs.begin());
    }
    while (!globals.empty() && min.end >= globals.front().end) {
        SPDLOG_TRACE("Erasing globals entry");
        globals.erase(globals.begin());
    }
    while (!macros.empty() && min.end >= macros.front().end) {
        SPDLOG_TRACE("Erasing macros entry");
        macros.erase(macros.begin());
    }
    while (!pragmas.empty() && min.end >= pragmas.front().end) {
        SPDLOG_TRACE("Erasing pragmas entry");
        pragmas.erase(pragmas.begin());
    }

    return min;
}

void CTUApexConsumer::HandleTranslationUnit(clang::ASTContext &Context) {
    const auto &translation_unit = Context.getTranslationUnitDecl();

    SPDLOG_INFO("Mapping functions to the call graph");
    FunctionFinderVisitor functionFinderVisitor(call_graph);
    functionFinderVisitor.TraverseTranslationUnitDecl(translation_unit);

    SPDLOG_INFO("Outlining Functions");
    outlineFunctions();

    SPDLOG_INFO("Outlining Globals");
    outlineGlobals();

    SPDLOG_INFO("Generate code skeleton");
    const auto &sm = Context.getSourceManager();
    std::string filename = sm.getFileEntryForID(sm.getMainFileID())->getName().str();
    filename.insert(filename.find_last_of('.') + 1, "mini.");
    std::ofstream outFile;
    outFile.open(filename);
    assert(outFile.is_open());
    while (!(includes.empty() && funcs.empty() && globals.empty() && macros.empty() && pragmas.empty())) {
        outFile << getNextExtractionBlock().contents << "\n";
    }

    outFile.close();
    SPDLOG_INFO("Generate wrapper calls");
    SPDLOG_CRITICAL("WRAPPER CALLS CAN NOT YET BE IMPLEMENTED");

    // Outline the code we want
    //SPDLOG_DEBUG("Generating outline");
    //assert(output.empty());
    //ASTOutliner outliner(&context, dump_anal_res);
    //auto extracted_source = outliner.generate_outline(marked, only_wrapper);

    //if (!only_wrapper) {
    //    auto main_source = outliner.generate_main_function(marked);
    //    output = extracted_source + main_source;
    //} else {
    //    output = extracted_source;
    //}

    SPDLOG_INFO("Clean CG for next run");
    for (const auto &node: call_graph->getNodes()) {
        if (node.second->has<ExtractionMetadata>()) {
            node.second->get<ExtractionMetadata>()->set_decl_node(nullptr);
        }
    }
}



