#
# File: TargetCollector.py
# This utility script allows to generate the whole-program call-graph for a given cmake target
# Script Version: 0.6.0
# License: Part of the MetaCG project. Licensed under BSD 3 clause license. See LICENSE.txt file at
# https://github.com/tudasc/metacg/LICENSE.txt
#

import argparse
import sys
import subprocess
import os
import json

if __name__ == '__main__':
    # Setup command line parsing
    # use -h to get pretty list
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        description='Generate Target specific callgraphs via cmake-file api')
    parser.add_argument('-g', '--generate',
                        help='Choose whether to generate the api files, the graph, or both',
                        required=False,
                        choices=["api", "graph", "both"], default="both")
    parser.add_argument('-b', '--build-directory', type=str, metavar="<path>", default=".",
                        help="The path to the cmake-build root folder", required=False)
    parser.add_argument('-e', '--extra-args', metavar="<list>", default="", help='pipe args to clang extra-arg',
                        type=str,
                        required=False)
    parser.add_argument('-o', '--output', metavar="<str>", default="wholeProgramCG.ipcg", type=str,
                        help='The name of the output file',
                        required=False)
    parser.add_argument('-a', '--cmake-args', metavar="<list>", help='pipe args to cmake', default="", type=str,
                        required=False)
    parser.add_argument('-t', '--target', metavar="<str>", type=str,
                        help='The target name for which to generate the callgraph',
                        required=True)
    parser.add_argument('--ci-concurrent-suffix', metavar="<str>", type=str,
                        help='Suffix appended to output filename in concurrent CI runs', required=False, default="")

    # parse all arguments after the first one, as this is the name we got called by
    parserObject: argparse.Namespace = parser.parse_args(sys.argv[1:])

    # if we want to generate a new api query file (or do both)
    if parserObject.generate in ['api', 'both']:
        # create the directories
        os.makedirs(parserObject.build_directory + "/.cmake/api/v1/query/", exist_ok=True)
        # create the file
        open(parserObject.build_directory + "/.cmake/api/v1/query/codemodel-v2", "a").close()
        # run cmake to generate the answers
        # we might need to pass additional parameters to cmake to match the original compile
        cmake_command: str = "cmake " + parserObject.cmake_args
        print("cd " + parserObject.build_directory + " && " + cmake_command)
        subprocess.run(cmake_command.split(" "), cwd=parserObject.build_directory)
    # if we want to generate a graph for the target
    if parserObject.generate in ['graph', 'both']:
        # we need to find the answer file for our target
        # for this, we list all targets, and find ours, via its prefix
        # target names are unique, meaning we get THE target with [0]
        targetFile: str = [f for f in os.listdir(parserObject.build_directory + "/.cmake/api/v1/reply") if
                           f.startswith("target-" + parserObject.target + "-")][0]

        # get the json from the file and extract the information we need
        # file is closed after json load finishes
        with open(parserObject.build_directory + "/.cmake/api/v1/reply/" + targetFile, 'r') as targetDescription:
            jsonFile: json = json.load(targetDescription)

        # get all source files from the compile-sources
        # we are not interested in generating a separate callgraph for *.h files,
        # as they are included anyway
        sources = [dict(elem)["path"] for elem in jsonFile["sources"] if not dict(elem)["path"].endswith("h")]

        includeSet: set[str] = set()

        for source in sources:
            with open(source, 'r') as sourceFile:
                for line in sourceFile:
                    if line.startswith("#include") and (line.endswith(".h\"\n") or line.endswith(".hpp\"\n") or line.endswith(".tpp\"\n") or line.endswith(".h>\n") or line.endswith(".hpp>\n") or line.endswith(".tpp>\n")):
                        includeSet.add(line)

        with open(parserObject.output, 'w') as outputFile:
            outputFile.write("".join([elem for elem in includeSet]))
            for source in sources:
                with open(source, 'r') as sourceFile:
                    for line in sourceFile:
                        if line.startswith("#include") and (line.endswith(".h\"\n") or line.endswith(".hpp\"\n") or line.endswith(".tpp\"\n") or line.endswith(".h>\n") or line.endswith(".hpp>\n") or line.endswith(".tpp>\n")  or line.endswith(">\n")):
                            continue
                        outputFile.write(line)
